function AlgParams = alg_SVMlight_selectParams(AlgParams, Inputs, Labels, verbose)
% function [AlgParams] = alg_SVMlight_selectParams(AlgParams, Inputs, Labels, [verbose=0])
%
% Selects SVM parameters (C, gamma) using validation
if nargin < 4, verbose = 1; end

N = size(Inputs, 1);
D = size(Inputs, 2);

ParamSelectionInfo = struct;

switch AlgParams.KernelType
    case 'linear'
        %     % coarse search -- use few data
        %     [I_train, I_valid] = sampleByClass(Labels, 0.25);
        %     C = [1e-4 1e-2 1e+0 1e+2 1e+4 1e+6 1e+8];
        %     E_valid = -1 .* ones(size(C));
        %     if verbose >= 1, disp(['Coarse search for linear SVM parameters...']); end
        %     [C_min, E_min] = iterate_linear(C, E_valid, AlgParams, Inputs, Labels, I_train, I_valid, verbose);
        %     
        %     % finer search -- use more data
        %     if verbose >= 1, disp(['Refining search around C=', num2str(C_min), ' ...']); end
        %     [I_train_fine, I_valid_fine] = sampleByClass(Labels, 0.75);
        %     C = C_min .* [1e-1 5e-1 1e+0 5e+1 1e+1];
        %     E_valid = -1 .* ones(size(C));
        %     % E_valid(find(C == C_min)) = E_min; now we change data size
        %     [C_min, E_min] = iterate_linear(C, E_valid, AlgParams, Inputs, Labels, I_train_fine, I_valid_fine, verbose);
        
        % coarse search -- 2cv
        numfolds = 2;
        indices = (1:N)'; % 'all data
        Folds = sampleFoldsByClass(indices, Labels(indices), numfolds);
        %C = 2.^(-5:2:15);
        C = 10.^(1:2:7);
        ParamSelectionInfo.Coarse_CRange = C;
        
        if verbose >= 1, 
            disp(['Coarse search for Linear SVM parameters...']);
        end
        [C_min, E_min, E_valid] = iterate_linear(C, AlgParams, Inputs, Labels, Folds, verbose);
        ParamSelectionInfo.Coarse_Errors = E_valid;
        
        % finer search -- 5cv
        numfolds = 5;
        indices = (1:N)'; % 'all data
        Folds = sampleFoldsByClass(indices, Labels(indices), numfolds);
        if verbose >= 1, 
            disp(['Refining search around C=', num2str(C_min), ' ...']);
        end        
        C = C_min .* (10.^(-1:1:1));
        ParamSelectionInfo.Fine_CRange = C;
        % E_valid(find(C == C_min)) = E_min; now we change data size
        [C_min, E_min, E_valid] = iterate_linear(C, AlgParams, Inputs, Labels, Folds, verbose);
        ParamSelectionInfo.Fine_Errors = E_valid;
        
        AlgParams.C = C_min;
        AlgParams.ParamSelectionInfo = ParamSelectionInfo;
        if verbose >= 1, 
            disp(['Selected C=', num2str(C_min), '.']); 
        end
        
    case 'rbf'
        % coarse search -- 2cv
        numfolds = 2;
        indices = (1:N)'; % all data
        Folds = sampleFoldsByClass(indices, Labels(indices), numfolds);
        if isequal(AlgParams.C, 'auto'), 
            C = 2.^(-5:2:15);
        else
            C = AlgParams.C;
        end;
        if isequal(AlgParams.KernelParam, 'auto'),
            Gamma = 2.^(-15:2:5);
        else
            Gamma = AlgParams.KernelParam;
        end;
        ParamSelectionInfo.Coarse_CRange = C;
        ParamSelectionInfo.Coarse_GammaRange = Gamma;
        if verbose >= 1, 
            disp(['Coarse search for RBF SVM parameters...']); 
        end
        [C_min, Gamma_min, E_min, E_valid] = iterate_rbf(C, Gamma, AlgParams, Inputs, Labels, Folds, verbose);
        ParamSelectionInfo.Coarse_Errors = E_valid;
        
        % finer search -- 5cv
        numfolds = 5;
        indices = (1:N)'; % all data
        Folds = sampleFoldsByClass(indices, Labels(indices), numfolds);
        if verbose >= 1, 
            disp(['Refining search around C=', num2str(C_min), ' Gamma=', num2str(Gamma_min), ' ...']); 
        end
        C = C_min .* (2.^(-2:1:2));
        Gamma = Gamma_min .* (2.^(-2:0.5:2));
        ParamSelectionInfo.Fine_CRange = C;
        ParamSelectionInfo.Fine_GammaRange = Gamma;
        % E_valid(find(C == C_min)) = E_min; now we change data size
        [C_min, Gamma_min, E_min, E_valid] = iterate_rbf(C, Gamma, AlgParams, Inputs, Labels, Folds, verbose);
        ParamSelectionInfo.Fine_Errors = E_valid;
        
        AlgParams.C = C_min;
        AlgParams.KernelParam = Gamma_min;
        AlgParams.ParamSelectionInfo = ParamSelectionInfo;
        if verbose >= 1, 
            disp(['Selected C=', num2str(C_min), ' Gamma=', num2str(Gamma_min), '.']); 
        end
        
    otherwise
        error('SVM parameter selection not implemented for this kernel');
end

return
%-------------------------------------------------------------------------
% function [C_min, E_min] = iterate_linear(C, E_valid, AlgParams, Inputs, Labels, I_train, I_valid, verbose)
% for i=1:length(C)
%     if E_valid(i) >= 0, continue; end;
%     AlgParams.C = C(i);
%     ModelParams = alg_SVMlight_learn(AlgParams, Inputs(I_train,:), Labels(I_train));
%     Outputs = alg_SVMlight_eval(ModelParams, Inputs(I_valid,:));
%     %E_valid(i) = sum(((Outputs - 0.5) .* (Labels(I_valid) - 0.5)) <= 0);
%     E_valid(i) = - sum(Labels(I_valid) .* log(Outputs) + (1 - Labels(I_valid)) .* log(1 - Outputs));
%     if verbose >= 1
%         disp(['Error for C', num2str(C(i)), ' = ', num2str(E_valid(i))]);
%     end
% end
function [C_min, E_min, E_valid] = iterate_linear(C, AlgParams, Inputs, Labels, Folds, verbose)
E_valid = zeros(length(C), 1);
for i_C = 1:length(C)
    %if E_valid(i) >= 0, continue; end;
    AlgParams.C = C(i_C);
    AlgParams.KernelParam = 0; % dummy
    for k = 1:length(Folds)
        ModelParams = alg_SVMlight_learn(AlgParams, Inputs(Folds{k}.ITrain,:), Labels(Folds{k}.ITrain));
        Outputs = alg_SVMlight_eval(ModelParams, Inputs(Folds{k}.IEval,:));
        %E_valid(i) = sum(((Outputs - 0.5) .* (Labels(I_valid) - 0.5)) <= 0);
        Error = - sum(Labels(Folds{k}.IEval) .* log(Outputs) + (1 - Labels(Folds{k}.IEval)) .* log(1 - Outputs));
        E_valid(i_C) = E_valid(i_C) + Error;
    end
    if verbose >= 1
        disp(['Error for C=', num2str(C(i_C)), ': ', num2str(E_valid(i_C))]);
    end
end
E_min = min(E_valid);
I_min = find(E_valid == E_min);
C_min = C(I_min(end));
return

%-------------------------------------------------------------------------
function [C_min, Gamma_min, E_min, E_valid] = iterate_rbf(C, Gamma, AlgParams, Inputs, Labels, Folds, verbose)
E_valid = zeros(length(C), length(Gamma));
for i_C = 1:length(C)
    for i_Gamma = 1:length(Gamma)
        %if E_valid(i) >= 0, continue; end;
        AlgParams.C = C(i_C);
        AlgParams.KernelParam = Gamma(i_Gamma);
        for k = 1:length(Folds)
            ModelParams = alg_SVMlight_learn(AlgParams, Inputs(Folds{k}.ITrain,:), Labels(Folds{k}.ITrain));
            Outputs = alg_SVMlight_eval(ModelParams, Inputs(Folds{k}.IEval,:));
            %E_valid(i) = sum(((Outputs - 0.5) .* (Labels(I_valid) - 0.5)) <= 0);
            Error = - sum(Labels(Folds{k}.IEval) .* log(Outputs) + (1 - Labels(Folds{k}.IEval)) .* log(1 - Outputs));
            E_valid(i_C, i_Gamma) = E_valid(i_C, i_Gamma) + Error;
        end
        if verbose >= 1
            disp(['Error for C=', num2str(C(i_C)), ' Gamma=', num2str(Gamma(i_Gamma)), ': ', num2str(E_valid(i_C, i_Gamma))]);
        end
    end
end
E_min = min(E_valid(:));
[I_min_C, I_min_Gamma] = find(E_valid == E_min);
C_min = C(I_min_C(end));
Gamma_min = Gamma(I_min_Gamma(end));
return
