function [Area, Sens, NSpec] = GetROC(Labels, Outputs, draw, saveFileName, saveFormat)
% [Area, Sens, NSpec] = GetROC(Labels, Outputs, draw, saveFileName)
%
% <Labels> and <Outputs> are column vectors.
% <draw> = 'draw' plots the ROC curve.
%   default is empty ('') value, which does not plot (except if <filename> given)
% if <draw> = 'draw' and <saveFileName> specified, the figure is PRINTed to file.
% <saveFormat> is 'png' by default is passed as '-d<saveFormat>' to PRINT.
% 'draw', '<filename>' will plot, save and keep the figure open.
% '', '<filename>' will plot, save and close the figure.
%
% <Area> is scalar area under the ROC curve.
% <Sens>, <NSpec> are col vectors of Sensitivity and 1-Specificity values respectively.

if nargin < 3, draw = ''; end;
if nargin < 4, saveFileName = ''; end;
if nargin < 5, saveFormat = 'png'; end;

J = find(~isnan(Outputs) & ~isnan(Labels));
Outputs = Outputs(J);
Labels = Labels(J);

[Outputs,I] = sort(Outputs);
Labels = Labels(I);

C = GetConfusion(Labels, Outputs, [0; Outputs; 1]);
CS = GetConfusionStats(C);

Sens = CS.Recall(:);
NSpec = 1 - CS.Specificity(:);

% sort by Sensitivity first
[Sens,I] = sort(Sens);
NSpec = NSpec(I);

% sort by 1-Specificity (stable sort)
[NSpec,I] = sort(NSpec);
Sens = Sens(I);

% calculate area under ROC curve
%Area = sum((NSpec(2:end) - NSpec(1:end-1)) .* (Sens(1:end-1) + Sens(2:end))) ./ 2;
% filter "unimportant" points for better approximation (no change in direction)
NSpec = [0; NSpec; 1];
Sens = [0; Sens; 1];


% we are really interested in the convex hull
Sens = [Sens ; 0];
NSpec = [NSpec ; 1];
I_convex = unique(convhull(NSpec, Sens));
NSpec  = NSpec(I_convex(1:end-1));
Sens  = Sens(I_convex(1:end-1));

Area = sum((NSpec(2:end) - NSpec(1:end-1)) .* (Sens(1:end-1) + Sens(2:end))) ./ 2;

if isequal(draw, 'draw') | ~isempty(saveFileName)
    figure;
    hold on;
    title(['ROC curve (area = ', num2str(Area) ,')']);
    ylabel('Sensitivity');
    xlabel('1 - Specificity');
    ylim([0 1.01]);
    H = plot(NSpec, Sens, 'k-');
    set(H, 'LineWidth', 1);
    
    if ~isempty(saveFileName)
        print(['-d', saveFormat], saveFileName);
    end
    
    if ~isequal(draw, 'draw')
        close;
    end
end
