function [bnet, Labels, Outputs] = tmp_combineBayesian(Labels, Outputs, NodeParents, ClassIDs, cov_prior, arrows)
% [bnet, Labels, Outputs, Conflicts] = tmp_combineBayesian(Labels, Outputs, NodeParents, ClassIDs, cov_prior, arrows)
% 
% Hierarchical Combining draft
%
% <Labels> is training labels (NumTrainingExamples-by-NumNodes)
%   OR filename pattern for label files to load. See loadLabels(...)
%   default if blank: use default in loadLabels(...)
% <Outputs> is classifier outputs (NumTrainingExamples-by-NumNodes)
%   OR filename pattern for Result files to load.
% <NodeParents> is indexed Node-Parent list (see loadNodeParents)
%   default if empty: NodeParents = loadNodeParents('', ClassIDs);
% <ClassIDs> is array of ClassIDs
%   default if empty: ClassIDs = getSelectedClassIDs;
% <cov_prior> is added to covariance priors. Default is zero.
% <arrows> is 'down' or 'up' or 'none', direction of the arrows.
%   default if empty: 'up'
%
% <bnet> is the resulting Bayes net.
% Optional return values <Labels> and <Outputs> are the loaded data if the
%  inputs were filename patterns. This is useful for avoiding reloading them
%  for subsequent calls.
% Optional return value <Conflicts> is column vector of hierarchical
%  conflict counts for each example.
%
% Example:
%  cd ~/research/function/rawresults
%  load svml_lin_Cinf
%  load svml_lin_Cinf_OutputsHeldOutRaw_withExcluded.mat
%  load NewLabels.mat.txt
%  bnet = tmp_combineBayesian(NewLabels, HeldOutRaw);

if nargin < 6, arrows = ''; end
if isempty(arrows) arrows = 'up'; end

if nargin < 5, cov_prior = 0; end

if nargin < 4, ClassIDs = []; end;

if nargin < 3, NodeParents = []; end;

if nargin < 1, Labels = []; end;
if isempty(Labels), Labels = ''; end;

if isempty(ClassIDs)
    ClassIDs = getSelectedClassIDs;
end

if isempty(NodeParents)
    % Load node parents (indexed)
    disp('Loading parents...');
    NodeParents = loadNodeParents('', ClassIDs);
end

if ischar(Labels)
    % load and convert to N1 labels (NaN -> 0) WHY?????????????????
    Labels = loadLabels(Labels, ClassIDs, 0);
end

if ischar(Outputs)
    disp('Loading outputs...');
    Outputs = loadOutputsFromResults(Outputs, size(Labels, 1), ClassIDs);
end

%% Get labels consistency (sanity check)
%disp('Analyzing labels consistency (sanity check)...');
%Conflicts = zeros(size(Labels,1), 1);
%% for each example
%for t = 1:size(Labels,1)
%    % for each parent-child pair
%    for j = 1:size(NodeParents, 1)
%        if Labels(t, NodeParents(j, 1)) == 0 && Labels(t, NodeParents(j, 2)) == 1
%            Conflicts(t,1) = Conflicts(t,1) + 1;
%        end
%    end
%end
%if sum(Conflicts) > 0
%    error('Sanity check failed: Inconsistent labels or parents')
%end


%% Get a consistency report for each example
%%disp('Analyzing consistency...');
%Conflicts = zeros(size(Labels,1), 1);
%% for each example
%for t = 1:size(Labels,1)
%    % for each parent-child pair
%    for j = 1:size(NodeParents, 1)
%        if Outputs(t, NodeParents(j, 1)) == 0 && Outputs(t, NodeParents(j, 2)) == 1
%            Conflicts(t,1) = Conflicts(t,1) + 1;
%        end
%    end
%end

% Create a Bayes net

disp('Creating DAG...');
D = size(Labels, 2);
N = size(Labels, 1);
dag = zeros(D, D);

% add label nodes (the first D nodes)
if isequal(arrows, 'down')
    for i = 1:length(NodeParents)
        dag(NodeParents(i,1), NodeParents(i,2)) = 1;
        %dag(NodeParents(i,2), NodeParents(i,1)) = 1; % upwards arrows
    end
elseif isequal(arrows, 'up')
    for i = 1:length(NodeParents)
        %dag(NodeParents(i,1), NodeParents(i,2)) = 1;
        dag(NodeParents(i,2), NodeParents(i,1)) = 1; % upwards arrows
    end
elseif isequal(arrows, 'none')
    %for i = 1:length(NodeParents)
        %dag(NodeParents(i,1), NodeParents(i,2)) = 1;
        %dag(NodeParents(i,2), NodeParents(i,1)) = 1; % upwards arrows
    %end
else
    error(['Invalid arrows parameter ', arrows]);
end

% sort here first
I_dag = topological_sort(dag);
sdag = dag(I_dag, I_dag);
dag = zeros(D*2, D*2);

dag(1:D,1:D) = sdag;
dag(1:D,(D+1):(D+D)) = eye(D);
% % TEST: add hierarchy among the g nodes too
% dag((D+1):(D+D),(D+1):(D+D)) = sdag;

% set node names
labelNames = cell(size(ClassIDs));
predNames = cell(size(ClassIDs));
for i=1:length(predNames)
    labelNames{i} = ['y ', ClassIDs{I_dag(i)}];
    predNames{i}  = ['g ', ClassIDs{I_dag(i)}];
end

disp('Creating Bayes net...');
% make Bayes net
bnet = mk_bnet(dag, [2 * ones(1,D), ones(1,D)], 'discrete', [1:D], 'names', [labelNames; predNames]);


% % learn probabilities
% for i = 1:D
%     bnet.CPD{i} = tabular_CPD(bnet, i);
%     bnet.CPD{i+D} = gaussian_CPD(bnet, i+D, 'cov_prior_weight', 0);
% end
% bnet = learn_params(bnet, [Labels(:,I_dag)+1, Outputs(:,I_dag)]');

% set probabilities
disp('Setting y CPDs...');
for i=1:D
    % label node CPTs
    parents = find(bnet.dag(:,i));
    numparents = length(parents);
    W = zeros(1, 2 .^ (numparents+1));
    IW = ind2subv(2 * ones(1,numparents+1), 1:length(W));
    parentLabels = Labels(:,I_dag(parents)) + 1;
    for k = 1:length(W)
        LI = find(sum(parentLabels == repmat(IW(k,1:end-1), N,1), 2) == numparents);
        if isempty(LI)
            W(k) = 0;
        else
            W(k) = sum(Labels(LI,I_dag(i))+1 == IW(k,end)) / length(LI);
        end
    end
    bnet.CPD{i} = tabular_CPD(bnet, i, 'CPT', W);
end

disp('Setting g CPDs...');
for i=1:D
    % prediction node CPTs
    g0 = Outputs(find(Labels(:,I_dag(i)) == 0), I_dag(i));
    g1 = Outputs(find(Labels(:,I_dag(i)) == 1), I_dag(i));
    
    % this was necessary in case some examples happened to be never held out
    g0 = g0(find(~isnan(g0)));
    g1 = g1(find(~isnan(g1)));
        
    bnet.CPD{i+D} = gaussian_CPD(bnet, i+D, 'mean', [mean(g0), mean(g1)], 'cov', [var(g0,1), var(g1,1)] + cov_prior);
    %disp([i, abs(mean(g0) - mean(g1)), std(g0,1), std(g1,1)]);
end

bnet.I_dag = I_dag;

disp('Done!');
return
