function [OutputsBayes] = combineBayesian_infer(bnet, Outputs, operation, numExamples, ClassIDs, fillNaNs)
% function [OutputsBayes] = combineBayesian_infer(bnet, Outputs, operation, numExamples, ClassIDs, fillNaNs)
%
% Makes Bayesian inference to resolve hierarchical inconsistencies.
%
% <bnet> is obtained from combineBayesian.
% <Outputs> is classifier predictions (NumExamples-by-NumNodes)
%   OR filename pattern for Result files (see loadOutputsFromResults)
% 
% <operation> is
%   'mpe' -- most probable explanation (discrete)
%   'marginal' -- marginal probabilities (continuous)
%   default if blank: 'marginal'
%
% <numExamples> and <ClassIDs> are only needed if <Outputs> is a filename pattern.
% 
% <fillNaNs> is
%   'fill' -- replace NaNs in the old outputs by normal random values. (default)
%   '' -- don't replace.

if nargin < 6, fillNaNs = 'fill'; end;

if nargin < 3, operation = ''; end;
if isempty(operation), operation = 'marginal'; end;

if ischar(Outputs)
    if nargin < 5, ClassIDs = []; end;
    if nargin < 4
        error('You must specify numExamples when Outputs is a filename pattern');
    end
    Outputs = loadOutputsFromResults(Outputs, numExamples, ClassIDs);
end

[N,D] = size(Outputs);

if isequal(fillNaNs, 'fill')
	% fill in NaN values in Outputs with normal random values
	for k=1:D
		I = find(isnan(Outputs(:,k)));
		J = find(~isnan(Outputs(:,k)));
		Mean = mean(Outputs(J,k));
		Stdev = std(Outputs(J,k));
		Outputs(I,k) = randn(size(I,1),1) .* Stdev + Mean;
	end
end


I_dag = bnet.I_dag;

% do inference
disp('Running inference...');
OutputsBayes = zeros(size(Outputs));

engine = jtree_inf_engine(bnet);
%engine = likelihood_weighting_inf_engine(bnet);

for t=1:N
    disp(t);
    evidence = cell(1,D*2);
    for i=1:D
    	v = Outputs(t,I_dag(i));
    	if ~isnan(v)
        	evidence{i+D} = v;
        end
    end
    [engine2, loglik] = enter_evidence(engine, evidence);

    if isequal(operation, 'marginal')
        % Marginal Probabilities
        for i=1:D
            marg = marginal_nodes(engine2, i);
            OutputsBayes(t,I_dag(i)) = marg.T(2);
            %disp([ClassIDs{I_dag(i)}, ': ', num2str(p)]);
        end
    elseif isequal(operation, 'mpe')        
        % Most Probable Explanation
        mpe = find_mpe(engine2, evidence);
        for i=1:D
            OutputsBayes(t,I_dag(i)) = mpe{i} - 1;
        end
    else
        error(['Undefined operation ', operation]);
    end
end

disp('Done!');
