function [HeldOutRaw, UnknownRaw, NotExcludedHeldOutRaw] = loadRawOutputsFromResults(pattern, numHeldOuts, numUnknowns, GOIDs)
% function [HeldOutRaw, UnknownRaw] = loadRawOutputsFromResults(pattern, numHeldOuts, numUnknowns, GOIDs)
%
% Loads held-out outputs of classifiers from Results.
%
% <pattern> is filename pattern with %s for GOID number.
%   default if empty: '~/research/function/data/outputs/outputs_%s.mat.txt'
% <numHeldOuts> is the number of training examples.
% <numUnknowns> is the number of unknown examples.
% <GOIDs> is array of GOIDs to load labels for
%   default if empty: GOIDs = getSelectedGOIDs;
%
% <HeldOutRaw> is unsigmoided held-out outputs (NumTrainingExamples-by-NumGOIDs)
%  If an example was not I_rows or held-out in all folds
%
% EXAMPLE:
%  [HeldOutRaw, UnknownRaw, NotExcludedHeldOutRaw] = loadRawOutputsFromResults(...
%    '~/research/function/rawresults/svml_lin_Cinf_V6_pred/%s_SVMlight_bagging10x1_linear_CInf.res.mat', 3465, 1059)

if nargin < 4, GOIDs = []; end;

if nargin < 3
    numUnknowns = 0;
end

if nargin < 2
    numHeldOuts = 0;
end

if numUnknowns == 0
    warning('You did not specify numUnknowns -- Unknowns will not be loaded.');
end

if numHeldOuts == 0
    warning('You did not specify numHeldOuts -- Unknowns will not be loaded.');
end

if numUnknowns == 0 && numHeldOuts == 0
	error('Neither numHeldOuts nor numUnknowns specified!');
end	

if isempty(pattern)
    error('You must specify Result file pattern');
end

if isempty(GOIDs)
    GOIDs = getSelectedGOIDs;
end


NotExcludedHeldOutRaw = ones(numHeldOuts, length(GOIDs)) .* NaN;
HeldOutRaw = ones(numHeldOuts, length(GOIDs)) .* NaN;
UnknownRaw = ones(numUnknowns, length(GOIDs)) .* NaN;
	
for i = 1:length(GOIDs)
    disp(num2str(i));
    goid = GOIDs{i};
    resultsfile = sprintf(pattern, goid(4:end));
    R = load(resultsfile);

	if numHeldOuts > 0
	    NotExcludedHeldOutRaw(R.Results.I_rows,i) = R.Results.HeldOutMedianOutputsRaw;
	
		% UPDATE: HeldOutRaw now includes excluded (N2) gene predictions instead of NaNs.
	    HeldOutRaw(R.Results.I_rows,i) = R.Results.HeldOutMedianOutputsRaw;
	    HeldOutRaw(R.Results.I_excludedRows,i) = R.Results.ExcludedMedianOutputsRaw;
    end

	if numUnknowns > 0
	    UnknownRaw(:,i) = R.Results.UnknownMedianOutputsRaw;
	end
end
