function Results = exp_run(AlgName, AlgParams, SInputs, Labels, aggregation, aggrParams, verbose, prenormalize, selectParams)
% function Results = exp_run(AlgName, AlgParams, SInputs, Labels, aggregation, aggrParams, verbose, prenormalize, selectParams)
% AlgName: string
%   The following functions must exist:
%     [ModelParams] = alg_<AlgName>_learn(AlgParams, Inputs, Labels)
%     [Outputs] = alg_<AlgName>_eval(ModelParams, Inputs)
% AlgParams: generic (usually struct) black-box
% SInputs: NxD possibly sparse matrix of training inputs in rows; no index row or col, no bias.
% SLabels: Nx1 possibly sparse col vector of training labels
% aggregation: string; 'cv' | 'bagging'| 'loo' (leave-one-out)
% aggrParams: [1x2 double];
%    [times folds] for 'cv'
%    [times ratio] for 'bagging'
%    [] for 'loo'
% verbose: integer; Default=0(silent)
% prenormalize: {'no'} | 'normalize' -- pre-normalize inputs
% selectParams: {'no'} | 'select' -- run parameter selection

clkBegin = clock;
rand('state',sum(100*clock));

% [TODO] parameters: read with defaults
if (nargin < 6) error('Required parameters missing'); end
%if ((nargin < 4) | isempty(inputsfile)), inputsfile = 'C:\Documents and Settings\zbarutcu\My Documents\Machine Learning\Olga\Data\ALL_inputs.mat.txt'; end
%if ((nargin < 5) | isempty(labelsfile)), labelsfile = 'C:\Documents and Settings\zbarutcu\My Documents\Machine Learning\Olga\Data\ALL_outputs.mat.txt'; end
%if ((nargin < 5) | isempty(folds)), folds = [3 5]; end
if ((nargin < 7) | isempty(verbose)), verbose = 0; end
if ((nargin < 8) | isempty(prenormalize)), prenormalize = 'no'; end;
if ((nargin < 9) | isempty(selectParams)), selectParams = 'no'; end;

% [TODO] parameters: validation and other processing
if ~ischar(AlgName), error('Algorithm name must be string'); end
funcLearn = inline(['alg_',AlgName,'_learn(AlgParams, Inputs, Labels)'],'AlgParams', 'Inputs', 'Labels');
funcEval = inline(['alg_',AlgName,'_eval(ModelParams, Inputs)'], 'ModelParams', 'Inputs');
if ~ischar(aggregation), error('Aggregation parameter must be string'); end
if (~isequal(aggregation, 'loo') && ~isequal(size(aggrParams), [1 2])), error(['Incorrect aggregation parameters size: ', mat2str(size(aggrParams))]); end;
%if ~ischar(inputsfile), error('Inputs file name must be string'); end
%if ~ischar(labelsfile), error('Labels file name must be string'); end
if ~(isnumeric(verbose) & isequal(size(verbose), [1 1])), error('Verbosity must be integer'); end

% Display parameters (verbose=1)
if verbose >= 1,
%    disp(['Inputs: ', inputsfile]);
%    disp(['Labels: ', labelsfile]);
    disp(['Inputs size: ', mat2str(size(SInputs))]);
    disp(['Labels size: ', mat2str(size(Labels))]);
    disp(['Aggregation: ', aggregation]);
    disp(['Aggr.Params: ', mat2str(aggrParams)]);
    disp(AlgParams);
end

Results = struct;

% don't use columns with < 2 nonzeros
Results.I_cols = find(sum((SInputs ~= 0),1) >= 2);

% pre-normalize inputs if requested
if isequal(prenormalize, 'normalize')    
    disp(['Pre-normalizing inputs...']);
    [Inputs, Results.ScalingMean, Results.ScalingStd] = normalizeInputs(SInputs(:, Results.I_cols));
else
    %Inputs = full(SInputs(:, Results.I_cols));
    % REUTERS: Not enough memory -- Use SInputs
    %Inputs = SInputs(:, Results.I_cols);
end
% REUTERS: save memory
%Labels = full(SLabels);
if verbose >= 1,
    disp(['Inputs filtered size: ', mat2str([size(SInputs, 1), length(Results.I_cols)])]);
end

% Validate dataset sizes
if (size(Labels,1) ~= size(SInputs,1)), error('Input rows do not match labels'); end
if (size(Labels,2) ~= 1), error('Labels must be a vector'); end
N = size(SInputs, 1);
D = length(Results.I_cols);

% prepare for class sampling, since positives are rare
I_pos = find(Labels);
I_neg = find(Labels == 0);
N_pos = size(I_pos,1);
N_neg = size(I_neg,1);
if (N_pos + N_neg ~= N), error('BUG: Negative and positive counts do not add up'); end;

% Initialize Results
Results.AlgParams = AlgParams;
Results.Labels = Labels;
Results.Aggregation = aggregation;
Results.AggregationParams = aggrParams;

if (isequal(aggregation,'cv'))
    times = aggrParams(1);
    folds = aggrParams(2);
    if (~isnumeric(times)) error('C.v. times must be integer'); end
    if (~isnumeric(folds)) error('C.v. folds must be integer'); end
    if (times < 1) error('C.v. times must be positive'); end
    if (folds < 2) error('C.v. folds must be >=2'); end
elseif (isequal(aggregation,'bagging'))
    times = aggrParams(1);
    ratio = aggrParams(2);
    if (~isnumeric(times)) error('Bagging times must be integer'); end
    if (~(ratio > 0 & ratio <= 1)) error('Bagging ratio must be in (0,1]'); end
    
    % for compatibility with C.v. related code
    folds = 1;

    % compute class training sizes (these are only for bagging)
    N_train_pos = floor(N_pos * ratio);
    N_train_neg = floor(N_neg * ratio);
    if (N_train_pos == 0) N_train_pos = 1; end;
    if (N_train_neg == 0) N_train_neg = 1; end;
    if (N_train_pos == N_pos) N_train_pos = N_pos-1; end;
    if (N_train_neg == N_neg) N_train_neg = N_neg-1; end;
elseif isequal(aggregation,'loo')
    % leave-one-out cross-validation
    folds = 1;
    times = 1;
else
    error(['Unknown aggregation: ', aggregation]);
end

Results.Folds = cell(times,folds);
Results.TestOutputs = zeros(N,times);
Results.TestConfusion = zeros(2,2);
Results.TrainingConfusion = zeros(2,2);

for t=1:times
    Perm_pos = randperm(N_pos)';
    Perm_neg = randperm(N_neg)';
    for f=1:folds
        clkFold = clock;
        Fold = struct;
        if (isequal(aggregation,'cv'))
            % compute folds in each class
            i_pos = floor((f-1)*N_pos/folds)+1;
            i_neg = floor((f-1)*N_neg/folds)+1;
            j_pos = floor(f*N_pos/folds);
            j_neg = floor(f*N_neg/folds);
            I_eval_pos = I_pos(Perm_pos(i_pos:j_pos));
            I_eval_neg = I_neg(Perm_neg(i_neg:j_neg));
            I_train_pos = I_pos(Perm_pos([1:(i_pos-1), (j_pos+1):N_pos]));
            I_train_neg = I_neg(Perm_neg([1:(i_neg-1), (j_neg+1):N_neg]));
            I_eval = [I_eval_pos ; I_eval_neg];
            I_train = [I_train_pos ; I_train_neg];
        elseif (isequal(aggregation,'bagging'))
            I_train_pos = I_pos(floor(rand(N_train_pos, 1) * N_pos + 1));
            I_train_neg = I_neg(floor(rand(N_train_neg, 1) * N_neg + 1));
            
            I_train = [I_train_pos ; I_train_neg];
            I_eval = setdiff((1:N)', unique(I_train));
        elseif isequal(aggregation,'loo')
            % leave-one-out cross-validation
            I_train = (1:N)';
            I_eval = (1:N)';
        else
            error(['Unknown aggregation: ', aggregation]);
        end
        % reshuffle just in case the algorithm doesn't
        I_eval = I_eval(randperm(size(I_eval,1)));
        I_train = I_train(randperm(size(I_train,1)));
        Fold.ITrain = I_train;
        Fold.ITest = I_eval;
%         size(I_train)
%         size(I_eval)
        
        % run parameter selection if requested, ON FIRST FOLD'S TRAINING SET
        if (t==1) && (f==1)
            if isequal(selectParams, 'select')
                disp('Running parameter selection...');
                funcSelectParams = inline(['alg_',AlgName,'_selectParams(AlgParams, Inputs, Labels)'], 'AlgParams', 'Inputs', 'Labels');
                AlgParams = feval(funcSelectParams, AlgParams, SInputs(I_train,Results.I_cols), Labels(I_train,:));
            else
                AlgParams.ParamSelectionInfo = [];
            end
            Results.AlgParams = AlgParams; % store in Results
        end

        % train
        if (verbose >=1) disp(['Fold ', num2str((t-1)*folds+f), ' training...']); end;
        clkTrain = clock;
        % REUTERS: directly ref SInputs for memory efficiency
        Fold.ModelParams = feval(funcLearn, AlgParams, SInputs(I_train,Results.I_cols), Labels(I_train,:));
        Fold.TrainingTime = etime(clock, clkTrain);
        if (verbose >=1) disp(['Training time: ', num2str(Fold.TrainingTime), ' secs']); end;
        
        % evaluate on training data
        clkEvalTra = clock;
        if (verbose >=1) disp(['Fold ', num2str((t-1)*folds+f), ' evaluating tra. outputs...']); end;
        %if isequal(aggregation,'loo') 
            Fold.ModelParams.LeaveOneOutNow = 1; 
        %end;
        % REUTERS: directly ref SInputs for memory efficiency
        Fold.TrainingOutputs = feval(funcEval, Fold.ModelParams, SInputs(I_train,Results.I_cols));
        %if isequal(aggregation,'loo') 
            Fold.ModelParams.LeaveOneOutNow = 0; 
        %end;
        if (verbose >=1) disp(['Tra. Evaluation time: ', num2str(etime(clock, clkEvalTra)), ' secs']); end;

        % get training confusion
        Fold.TrainingConfusion = GetConfusion(Labels(I_train,:), Fold.TrainingOutputs, [0.5]);

        if isequal(aggregation,'loo')
            Fold.TestOutputs = Fold.TrainingOutputs;
            Fold.TestConfusion = Fold.TrainingConfusion;
        else
            % evaluate on test data
            clkEvalTest = clock;
            %Inputs_subset = ;  % reuse variable for memory efficiency
            if (verbose >=1) disp(['Fold ', num2str((t-1)*folds+f), ' evaluating test outputs...']); end;
            Fold.TestOutputs = feval(funcEval, Fold.ModelParams, SInputs(I_eval,Results.I_cols)); % REUTERS: Ref SInputs to save mem
            if (verbose >=1) disp(['Test Evaluation time: ', num2str(etime(clock, clkEvalTest)), ' secs']); end;

            % get test confusion
            Fold.TestConfusion = GetConfusion(Labels(I_eval,:), Fold.TestOutputs, [0.5]);
        end
        
        if (verbose >=2)
            disp('Tra. Confusion:');
            DisplayConfusion(Fold.TrainingConfusion);
            disp('Test Confusion:');
            DisplayConfusion(Fold.TestConfusion);
            disp(GetConfusionStats(Fold.TestConfusion));
        end
        
        % update summary results
        Results.TestOutputs(I_eval,t) = Fold.TestOutputs;
        Results.TestConfusion = Results.TestConfusion + Fold.TestConfusion;
        Results.TrainingConfusion = Results.TrainingConfusion + Fold.TrainingConfusion;
        
        % store fold results
        Fold.TotalTime = etime(clock, clkFold);
        Results.Folds{t,f} = Fold;
    end
end

% update final summary results
Results.TestConfusion = Results.TestConfusion ./ times;
if (folds == 1) % e.g. bagging or loo
    Results.TrainingConfusion = Results.TrainingConfusion ./ times;
else
    Results.TrainingConfusion = Results.TrainingConfusion ./ (times * (folds-1));
end

% Post-Process: make combined predictions on held-out examples
if isequal(aggregation,'loo')
    Results.HeldOutMeanOutputs = Results.Folds{1}.TestOutputs;
    Results.HeldOutMedianOutputs = Results.Folds{1}.TestOutputs;
else
    disp('Predicting Held-outs (WHY????)');
    [MeanY, MedianY] = predictHeldOuts(funcEval, Results, SInputs(:, Results.I_cols));
    Results.HeldOutMeanOutputs = MeanY;
    Results.HeldOutMedianOutputs = MedianY;
end

Results.TotalTime = etime(clock, clkBegin);

if (verbose >= 1)
    disp(['Total time = ', num2str(Results.TotalTime), ' secs.']);
    disp('Mean Tra. Confusion:');
    DisplayConfusion(Results.TrainingConfusion);
    disp('Mean Test Confusion:');
    DisplayConfusion(Results.TestConfusion);
    disp(GetConfusionStats(Results.TestConfusion));
    disp('(End of c.v.)');
end

return

%-------------------------------
function DisplayConfusion(C)
% display confusion in a friendly format
TN = C(1,1); FP = C(1,2);
FN = C(2,1); TP = C(2,2);
disp(sprintf('  TN=%4d ; FP=%4d', TN, FP));
disp(sprintf('  FN=%4d ; TP=%4d', FN, TP));
return
