function writeWebData_Compare(ResArrayN1,ResArrayN2,NodeInfos)
% function writeWebData_Compare(ResArrayN1,ResArrayN2,NodeInfos)
%
% Writes tabulated data for web page.
% !!!NOTE: This compares two sets of raw (unBayesed) results.
%
% <ResArrayN1> and <ResArrayN2> are 
%      cell-arrays of Results 
%   OR filename templates for result files.
% <NodeInfos> is (optional)
%      cell-array of NodeInfo
%   OR filename param for loadNodeInfos
%   OR nothing -- uses loadNodeInfos() default.
% 

if nargin < 3, NodeInfos = []; end;

if isempty(NodeInfos)
    NodeInfos = loadNodeInfos;
elseif ischar(NodeInfos)
    NodeInfos = loadNodeInfos(NodeInfos);
end

for iInfo=1:size(NodeInfos.GOIDs,1)
    GOID = NodeInfos.GOIDs{iInfo};

    ResultsN1 = FindResults(ResArrayN1, GOID, 'Left-side');
    if isempty(ResultsN1)
        warning(['Left-side results for ', GOID, ' not found!']);
        continue;
    end 
    ResultsN2 = FindResults(ResArrayN2, GOID, 'Right-side');
    if isempty(ResultsN2)
        warning(['Right-side results for ', GOID, ' not found!']);
        continue;
    end 
    
    % stats
    NPos = sum(ResultsN1.Labels,1);
    NNegDiff = size(ResultsN1.Labels,1)-size(ResultsN2.Labels,1);
    N1N2adjustment = sum(ResultsN1.Labels == 0) ./ sum(ResultsN2.Labels == 0);
    thresholds = 0:0.01:1;
    ConfN1 = GetConfusion(ResultsN1.Labels, ResultsN1.HeldOutMedianOutputs, thresholds);
    ConfN2 = GetConfusion(ResultsN2.Labels, ResultsN2.HeldOutMedianOutputs, thresholds);
    StatsN1 = GetConfusionStats(ConfN1);
    StatsN2 = GetConfusionStats(ConfN2, N1N2adjustment);
    
    % find best prec and its rec for rec>=.10, rec>=.25, rec>=10/NPos
    [N1_010_Prec, N1_010_Rec] = GetPrecRec(StatsN1, 0.10);
    [N2_010_Prec, N2_010_Rec] = GetPrecRec(StatsN2, 0.10);
    [N1_025_Prec, N1_025_Rec] = GetPrecRec(StatsN1, 0.25);
    [N2_025_Prec, N2_025_Rec] = GetPrecRec(StatsN2, 0.25);
    [N1_10P_Prec, N1_10P_Rec] = GetPrecRec(StatsN1, 10/NPos);
    [N2_10P_Prec, N2_10P_Rec] = GetPrecRec(StatsN2, 10/NPos);

    % write out
    disp(['"', GOID(4:end), '"', ...
        '|', num2str(NPos), ...
        '|', num2str(NNegDiff), ...
        '|', num2str(mean(ResultsN1.AUCs)), ...
        '|', num2str(std(ResultsN1.AUCs)), ...
        '|', num2str(ResultsN1.HeldOutMedianAUC), ...
        '|', num2str(mean(ResultsN2.AUCs)), ...
        '|', num2str(std(ResultsN2.AUCs)), ...
        '|', num2str(ResultsN2.HeldOutMedianAUC), ...
        '|', num2str(N1_010_Prec), '|', num2str(N1_010_Rec), ...
        '|', num2str(N2_010_Prec), '|', num2str(N2_010_Rec), ...
        '|', num2str(N1_025_Prec), '|', num2str(N1_025_Rec), ...
        '|', num2str(N2_025_Prec), '|', num2str(N2_025_Rec), ...
        '|', num2str(N1_10P_Prec), '|', num2str(N1_10P_Rec), ...
        '|', num2str(N2_10P_Prec), '|', num2str(N2_10P_Rec), ...
        '|', '"', NodeInfos.nodeNames{iInfo}, '"', ...
        ]);
end

%----------------------------
function [prec, rec] = GetPrecRec(Stats, reclevel)
if reclevel > 1, reclevel = 1; end
prec = max(Stats.Precision(find(Stats.Recall >= reclevel)));
if isempty(prec), 
    prec = 0;
    rec = reclevel;
else
    rec = max(Stats.Recall(find(Stats.Precision == prec)));
end
return

%----------------------------
function Results = FindResults(ResArray, GOID, name)
if iscell(ResArray)
    % match N1 result
    i1=0;
    for j=1:size(ResArray,1)
        if isequal(GOID, ResArray{j}.GOID)
            i1 = j;
            break
        end
    end
    if i1 == 0
        Results = [];
        return;
    end
    Results = ResArray{i1};
elseif ischar(ResArray)
    filename = sprintf(ResArray, GOID(4:end));
    Results = load(filename);
    Results = Results.Results;
else
    error(['Invalid parameter for ', name, ' results']);
end
return
