#!/usr/bin/python
"""
GOTermFinder_export - Exports GO annotations as a geneORF-GOID pairs.
by Zafer Barutcuoglu <zbarutcu@princeton.edu> 2004/11/29
Copyright (c) 2004 Princeton University

USAGE: GOTermFinder_export <options>
  Options:
    -h       | --help                - Display this information and exit.
    -v       | --verbose             - Display diagnostics (to stderr). Recommended.
    -g<file> | -go-file=<file>
    -G<url>  | -go-url=<url>
    -a<file> | -annot-file=<file>
    -A<url>  | -annot-url=<url>
    -l<file> | -genelist-file=<file>
    -L<url>  | -genelist-url=<url>
    -o<file> | -output-pairs=<file>
    -n<file> | -output-nodes=<file>

Each URL parameter can be specified as a period (e.g. "-G.") to use the default URL.

There are three inputs, with default URLs given below:

 <genelist>
  URL: ftp://genome-ftp.stanford.edu/pub/yeast/data_download/chromosomal_feature/SGD_features.tab
  This file is used for CommonGeneName-to-ORF mapping, and if omitted, pairs may contain non-ORF names.
  Although technically optional, it should always be provided for all-ORF output.
  Any gene not in this list is not included in the pairs, with a warning in verbose output.

 <go>
  URL: http://www.geneontology.org/ontology/process.ontology
  This file contains the GO hierarchy nodes.
  It MUST be specified, as a file or a URL.
  If both a file and a URL are specified, the file is overwritten from the URL.
  
 <annot>
  URL: http://www.geneontology.org/cgi-bin/GO/downloadGOGA.pl/gene_association.sgd.gz
  This file contains the gene annotations to GO nodes.
  It MUST be specified, as a file or a URL.
  If both a file and a URL are specified, the file is overwritten from the URL.

There are two possible output files:

 <pairs>
  This file contains lines of tab-separated ORF-nodeID pairs, including annotations propagated from child nodes.
  It can be omitted to not generate this output.
 <nodes>
  This file contains lines of tab-separated nodeID-nodeName pairs for all GO nodes.
  It can be omitted to not generate this output.

Omitting both output files is allowed, for example to only download and save input files.

Example:
  GOTermFinder_export -v -G. -A. -L. -oGTFpairs.txt -nGTFnodes.txt
   Verbosely download all files from default URLs (without saving) and output GTFpairs.txt and GTFnodes.txt.
"""

__version__ = "1.0"

import sys
import getopt

class UsageError(Exception):
    """Usage error exception class.
    
    Attributes:
        msg -- Error message
    """
    
    def __init__(self, msg):
        self.msg = msg
        
    def __str__(self):
        return self.msg


class ExitCodes:
    """A namespace class for exit status codes."""
    
    Success = 0
    ErrorsFound = 1
    Usage = 2
    Crash = 3

def printVerbose(s):
    sys.stderr.write(s);
    sys.stderr.flush();

def main(argv = None):
    """Main program body."""
    
    if argv is None:
        argv = sys.argv
    
    try:
        # parse command line into opts and args
        try:
            opts, args = getopt.getopt(argv[1:], "hvg:G:a:A:l:L:o:n:", [
                "help", "verbose", "go-file=", "go-url=", "annot-file=", "annot-url=",
                "genelist-file=", "genelist-url=", "output-pairs=", "output-nodes="
                ])
        except getopt.error, msg:
             raise UsageError(msg)
        
        # process options
        sOntologyFile = None
        sOntologyURL = None
        sAssocFile = None
        sAssocURL = None
        sIndexFile = None
        sIndexURL = None
        sOutputFile = None
        sOutputNodesFile = None
        bVerbose = False
        for o, a in opts:
            if o in ("-h", "--help"):
                print __doc__
                return ExitCodes.Success
            elif o in ("-v", "--verbose"):
                bVerbose = True
            elif o in ("-g", "--go-file"):
                sOntologyFile = a
            elif o in ("-G", "--go-url"):
                sOntologyURL = a
                if sOntologyURL == ".": sOntologyURL = 'default'
            elif o in ("-a", "--annot-file"):
                sAssocFile = a
            elif o in ("-A", "--annot-url"):
                sAssocURL = a
                if sAssocURL == ".": sAssocURL = 'default'
            elif o in ("-l", "--genelist-file"):
                sIndexFile = a
            elif o in ("-L", "--genelist-url"):
                sIndexURL = a
                if sIndexURL == ".": sIndexURL = 'default'
            elif o in ("-o", "--output-pairs"):
                sOutputFile = a
            elif o in ("-n", "--output-nodes"):
                sOutputNodesFile = a
        
        # process arguments
        if len(args) != 0:
            raise UsageError("Too many arguments!")
        
        # load GeneIndex
        import GeneIndex
        geneIndex = GeneIndex.GeneIndex()
        geneIndex.bVerbose = bVerbose
        if sIndexURL is None:
            if sIndexFile is not None and len(sIndexFile) > 0:
                if bVerbose: printVerbose('Loading gene list from "%s"\n' % sIndexFile)
                geneIndex.loadFile(sIndexFile, 'auto', bVerbose)
                if bVerbose: printVerbose('Gene list loaded.\n')
            else:
                printVerbose('WARNING: Gene list source not specified! CGN->ORF matching will be unreliable!\n')
                geneIndex = None
        else:
            if bVerbose: printVerbose('Downloading gene list from "%s"\n' % sIndexURL)
            if sIndexFile is not None and len(sIndexFile) > 0:
                if bVerbose: printVerbose(' and saving to file "%s"\n' % sIndexFile)
                geneIndex.loadURL(sIndexURL, 'auto', sIndexFile, bVerbose)
                if bVerbose: printVerbose('Gene list downloaded and saved.\n')
            else:
                geneIndex.loadURL(sIndexURL, 'auto', None, bVerbose)
                if bVerbose: printVerbose('Gene list downloaded (not saved).\n')
        
        # load Ontology
        import ontology
        go = ontology.Ontology()
        if sOntologyURL is None:
            if sOntologyFile is not None and len(sOntologyFile) > 0:
                if bVerbose: printVerbose('Loading GO nodes from "%s"\n' % sOntologyFile)
                go.loadFile(sOntologyFile, 'auto', bVerbose)
                if bVerbose: printVerbose('GO nodes loaded.\n')
            else:
                raise UsageError('GO nodes file or URL source not specified')
        else:
            if bVerbose: printVerbose('Downloading GO nodes from "%s"\n' % sOntologyURL)
            if sOntologyFile is not None and len(sOntologyFile) > 0:
                if bVerbose: printVerbose(' and saving to file "%s"\n' % sOntologyFile)
                go.loadURL(sOntologyURL, 'auto', sOntologyFile, bVerbose)
                if bVerbose: printVerbose('GO nodes downloaded and saved.\n')
            else:
                go.loadURL(sOntologyURL, 'auto', None, bVerbose)
                if bVerbose: printVerbose('GO nodes downloaded (not saved).\n')
        
        # load Assocation
        import association
        assoc = association.Association(None, geneIndex.Any2ORF)
        assoc.bVerbose = bVerbose
        if sAssocURL is None:
            if sAssocFile is not None and len(sAssocFile) > 0:
                if bVerbose: printVerbose('Loading annotations from "%s"\n' % sAssocFile)
                assoc.loadFile(sAssocFile, 'auto', bVerbose)
                if bVerbose: printVerbose('Annotations loaded.\n')
            else:
                raise UsageError('Annotations file or URL source not specified')
        else:
            if bVerbose: printVerbose('Downloading annotations from "%s"\n' % sAssocURL)
            if sAssocFile is not None and len(sAssocFile) > 0:
                if bVerbose: printVerbose(' and saving to file "%s"\n' % sAssocFile)
                assoc.loadURL(sAssocURL, 'auto', sAssocFile, bVerbose)
                if bVerbose: printVerbose('Annotations downloaded and saved.\n')
            else:
                assoc.loadURL(sAssocURL, 'auto', None, bVerbose)
                if bVerbose: printVerbose('Annotations downloaded (not saved).\n')
        
        # add annotations
        if bVerbose: printVerbose('Mapping annotations to ontology...\n')
        go.addAnnotations(assoc, bVerbose)
        if bVerbose: printVerbose('Annotations mapped.\n')
        
        # output pairs
        if sOutputFile is not None and len(sOutputFile) > 0:
            if bVerbose: printVerbose('Preparing pairs data...\n')
            sPairs = ''
            for n in go.nodes:
                for g in n.genes:
                    #sPairs = sPairs + ('%s\t%s\n' % (g['ORF'], '\t'.join(n.ids)))
                    sPairs = sPairs + ('%s\t%s\n' % (g['ORF'], n.ids[0]))
            if bVerbose: printVerbose('Saving pairs to file "%s"\n' % sOutputFile)
            f = file(sOutputFile, 'w')
            f.write(sPairs)
            f.close()
            if bVerbose: printVerbose('Pairs file saved (%d bytes).\n' % len(sPairs))
            sPairs = ''
        
        if sOutputNodesFile is not None and len(sOutputNodesFile) > 0:
            if bVerbose: printVerbose('Preparing nodes data...\n')
            sNodes = ''
            for n in go.nodes:
                sNodes = sNodes + ('%s\t%s\n' % (n.ids[0], n.name))
            
            if bVerbose: printVerbose('Saving nodes to file "%s"\n' % sOutputNodesFile)
            f = file(sOutputNodesFile, 'w')
            f.write(sNodes)
            f.close()
            if bVerbose: printVerbose('Nodes file saved (%d bytes).\n' % len(sNodes))
            sNodes = ''
        
        iExitCode = ExitCodes.Success
            
        if bVerbose: printVerbose('Exiting with success.\n')
        return iExitCode
        
        
##    except ParserException, pe:
##        print "%s (%d): %s" % (pe.fileName, pe.lineNumber, pe.description)
##        return ExitCodes.ErrorsFound
        
    except UsageError, err:
        print >>sys.stderr, "USAGE ERROR: %s" % err.msg
        print >>sys.stderr, "for help use --help"
        return ExitCodes.Usage
        
##    except:
##        print >>sys.stderr, "ERROR: %s" % sys.exc_info()[0]
##        #raise
##        return ExitCodes.Crash
        
##	except:
##		print >>sys.stderr, "UNKNOWN ERROR!"
##		return ExitCodes.Crash



if __name__ == "__main__":
    sys.exit(main())
