"""
Gene Interactions (GRID) data parser
Zafer Barutcuoglu 2003/11/25
This module parses the bioinformatical General Repository for Interaction Datasets (GRID) data.
It outputs a file for each experiment type, suitable for importing into Matlab as a sparse matrix.
"""

from TextData import TextData
import os

class Interactions(TextData):
    """GRID interaction data class.
    
    self.experiments is dict(sName->Experiment)
    self.genes is list of gene names for indexing
    self.sources is list of sources for indexing
    self.PubMed_IDs is list of PubMed IDs for indexing
    """

    #HEADER_ROW = 'ORF_A\tORF_B\tEXPERIMENTAL_SYSTEM\tSOURCE\tPUBMED_ID\n'
    HEADER_ROW = 'INTERACTOR_A\tINTERACTOR_B\tALIASES_FOR_A\tALIASES_FOR_B\tEXPERIMENTAL_SYSTEM\tSOURCE\tDIRECTION\tPUBMED_ID\r\n'
    NUM_FIELDS = len(HEADER_ROW.split('\t'))

    class Experiment:
        """Data for one type of experiment.
        self.data is dict: (A_index,B_index)->(source_index,PubMedIndices_list)
        """
        def __init__(self, name):
            
            self.name = name
            self.data = {}
        
        
        def addItem(self, A, B, source, pubmeds):
            self.data[(A, B)] = (source, pubmeds)
        
##        def getMatrix(self, orfList):
##            import numarray
##            N = len(orfList)
##            m = numarray.zeros([N,N], type = numarray.Int8)
##            for (a,b) in self.data.iterkeys():
##                if a not in orfList or b not in orfList:
##                    continue
##                r = orfList.index(a)
##                c = orfList.index(b)
##                m[r,c] = 1
##            return m
            
            
        def save(self, outfile, rowORFList, colORFList, indexList = 'default', transpose = False, rowoffset = 0, coloffset = 0):
            """ indexList: None or [] -> don't save indices; default -> colORFList
            """
            if indexList == 'default':
                indexList = colORFList
            elif indexList is None or indexList == []:
                indexList = None
            
            ftype = type(outfile)
            if ftype is str:
                f = file(outfile, 'wt')
            elif ftype is file:
                f = outfile
            
            try:
                # write indices
                if indexList is not None:
                    for i in range(0,len(colORFList)):
                        f.write('%d\t%d\t%d\n' % (1+rowoffset, i+1+coloffset, indexList.index(colORFList[i])))
                    rowoffset += 1

                # write items
                for (a, b) in self.data.iterkeys():
                    try:
                        if not transpose:
                            r = rowORFList.index(a)
                            c = colORFList.index(b)
                        else:
                            r = rowORFList.index(b)
                            c = colORFList.index(a)                            
                    except ValueError:
                        continue
                    f.write('%d\t%d\t1\n' % (r+1+rowoffset, c+1+coloffset))
                
                # ensure size by writing out item (m,n)
                if not self.data.has_key((len(rowORFList), len(colORFList))):
                    f.write('%d\t%d\t0\n' % (len(rowORFList)+rowoffset, len(colORFList)+coloffset))
                
            finally:
                if ftype is str:
                    f.close()
        
        def isTransitive(self):
            for (a,b) in self.data.iterkeys():
                for (c,d) in self.data.iterkeys():
                    if a == c and b != d \
                       and not self.data.has_key((b,d)) \
                       and not self.data.has_key((d,b)):
                        return False
                    elif b == d and a != c \
                       and not self.data.has_key((a,c)) \
                       and not self.data.has_key((c,a)):
                        return False
                    elif a == d and b != c \
                       and not self.data.has_key((b,c)) \
                       and not self.data.has_key((c,b)):
                        return False
                    elif b == c and a != d \
                       and not self.data.has_key((a,d)) \
                       and not self.data.has_key((d,a)):
                        return False
            return True
        
        def countIntransitives(self):
            cnt = 0
            for (a,b) in self.data.iterkeys():
                for (c,d) in self.data.iterkeys():
                    if a == c and b != d \
                       and not self.data.has_key((b,d)) \
                       and not self.data.has_key((d,b)):
                        cnt += 1
                    elif b == d and a != c \
                       and not self.data.has_key((a,c)) \
                       and not self.data.has_key((c,a)):
                        cnt += 1
                    elif a == d and b != c \
                       and not self.data.has_key((b,c)) \
                       and not self.data.has_key((c,b)):
                        cnt += 1
                    elif b == c and a != d \
                       and not self.data.has_key((a,d)) \
                       and not self.data.has_key((d,a)):
                        cnt += 1
            return cnt
    
    
    def __init__(self, infile = None, translator = None):
        """Constructor.
        <file> may be either a file name, open file object, or None.
        <bSkipComments> skips lines until header row is found.
        """
        #self.DEFAULT_FILE_NAME = 'C:\\Documents and Settings\\Zafer\\My Documents\\Machine Learning\\Olga\\Data\\GRID\\interactions.txt'
        self.DEFAULT_FILE_NAME = os.getenv('HOME') + '/research/function/rawdata/GRID/interactions.txt'
        
        self.genes = []
        self.funcTranslateName = translator
        
        self.experiments = {}
        self.sources = []
        self.PubMed_IDs = []
        

    def loadLines(self, lines, bSkipComments = True):
        """Loads data from a list of data lines.
        <bSkipComments> skips lines until header row is found.
        Format: INTERACTOR_A\tINTERACTOR_B\tALIASES_FOR_A\tALIASES_FOR_B\tEXPERIMENTAL_SYSTEM\tSOURCE\tDIRECTION\tPUBMED_ID\r\n
        """
        iLen = len(lines)
        iCur = 0
        
        if bSkipComments:
            while iCur < iLen and lines[iCur] != self.HEADER_ROW:
                iCur += 1
            if iCur == iLen:
                raise Exception('Header row not found!')
            iCur += 1 # skip header row itself
        
        while iCur < iLen:
            fields = lines[iCur].split('\t')
            if len(fields) != self.NUM_FIELDS:
                raise Exception('Incorrect number of columns at line %d: %d expected, %d found!' % (iCur+1, self.NUM_FIELDS, len(fields)))
            
            ORF_A, ORF_B, aliases_A, aliasesB, EXPERIMENTAL_SYSTEM, SOURCE, DIRECTION, PUBMED_ID = fields
            # trim these, maybe?
            if self.funcTranslateName is not None:
                t = self.funcTranslateName(ORF_A)
                if t is None: print '*** WARNING: Gene lost in translation "%s"' % ORF_A ; continue
                else: ORF_A = t
                t = self.funcTranslateName(ORF_B)
                if t is None: print '*** WARNING: Gene lost in translation "%s"' % ORF_B ; continue
                else: ORF_B = t
            
            if ORF_A not in self.genes:
                self.genes.append(ORF_A)
            if ORF_B not in self.genes:
                self.genes.append(ORF_B)
            
            # source name to index
            if SOURCE not in self.sources:
                self.sources.append(SOURCE)
            source = self.sources.index(SOURCE)
            
            # PubMed ID names to indices
            PubMedIDList = [ID.strip() for ID in PUBMED_ID.split(';') if ID.strip() != '']
            PubMedIndices = []
            for ID in PubMedIDList:
                if ID not in self.PubMed_IDs:
                    self.PubMed_IDs.append(ID)
                PubMedIndices.append(self.PubMed_IDs.index(ID))
            
            # find or create experiment
            if not self.experiments.has_key(EXPERIMENTAL_SYSTEM):
                self.experiments[EXPERIMENTAL_SYSTEM] = self.Experiment(EXPERIMENTAL_SYSTEM)
            experiment = self.experiments[EXPERIMENTAL_SYSTEM]
            
            # add data
            experiment.addItem(ORF_A, ORF_B, source, PubMedIndices)
            if DIRECTION == 'BOTH':
                experiment.addItem(ORF_B, ORF_A, source, PubMedIndices)
            
            # next line
            iCur += 1

    
    
##    def saveAll(self, orfList, path, saveIndices = True):
##        if not path.endswith('\\') and not path.endswith('/'):
##            import os
##            path = path + os.sep
##            
##        for ex in self.experiments.itervalues():
##            ex.save(orfList, path + ex.name + '.txt', saveIndices)

    def saveMerged(self, outfile, rowORFList, colORFList, indexList = 'default', includeTransposes = True):
        """ indexList: None or [] -> don't save indices; default -> colORFList
        """
        if indexList == 'default':
            indexList = colORFList
        elif indexList is None or indexList == []:
            indexList = None
        
        ftype = type(outfile)
        if ftype is str:
            outfile = file(outfile, 'wt')
        
        try:
            coloffset = 0
            if indexList is not None:
                for i in range(0,len(rowORFList)):
                    # leave (0,0) zero, hence 2+i
                    outfile.write('%d\t%d\t%d\n' % (2+i, 1+coloffset, indexList.index(rowORFList[i])))
                coloffset += 1
            
            for ex in self.experiments.itervalues():
                ex.save(outfile, rowORFList, colORFList, indexList, transpose = False, coloffset = coloffset)
                coloffset += len(colORFList)
                if includeTransposes:
                    ex.save(outfile, rowORFList, colORFList, indexList, transpose = True, coloffset = coloffset)
                    coloffset += len(colORFList)
        finally:
            if ftype is str:
                outfile.close()
        
        return coloffset

    def printSummary(self):
        print '%s summary:' % self.__class__.__name__
        print ' Experiments (%d total):' % len(self.experiments)
        for e in self.experiments.itervalues():
            print '  %s (%d entries)' % (e.name, len(e.data))
        print ' Num. Genes = %d' % len(self.genes)
        print ' Num. Sources = %d' % len(self.sources)
        print ' Num. PubMed IDs = %d' % len(self.PubMed_IDs)
        print '(End)'
    
# Main test code
# grid = GRID.Interactions() ; grid.loadFile()
