"""
Gene Association data parser
Zafer Barutcuoglu 2003/11/25

This module parses the bioinformatical Gene Association data, which is the mapping of genes to Gene Ontology nodes.

The term "GOID" used throughout is short for Gene Ontology ID, of the format "GO:1234567".
"""

from TextData import TextData
from Table import Table

import sys
import os

class Association(TextData):
    """Gene Association data class.
    
    self.genes is Table(['SGDID','CGN','ORF','nodes'])
    """
    
    def __init__(self, infile = None, translator = None, ignoreNodes = []):
        """Constructor.
        <infile> may be either a file name, open file object, or None.
        """
        #self.DEFAULT_FILE_NAME = 'C:\\Documents and Settings\\Zafer\\My Documents\\Machine Learning\\Olga\\Data\\GO\\gene_association.sgd.txt'
        self.DEFAULT_FILE_NAME = os.getenv('HOME') + '/research/function/rawdata/GO/gene_association.sgd.txt'
        self.DEFAULT_URL = 'http://www.geneontology.org/cgi-bin/GO/downloadGOGA.pl/gene_association.sgd.gz'

        self.genes = Table(['SGDID','CGN','ORF','nodes', 'directNodes'])
        self.funcTranslateORF = translator
        
        self.ignoreNodes = ignoreNodes
        if self.ignoreNodes is None: self.ignoreNodes = []

        self.bVerbose = True
        
        if infile is not None:
            self.loadFile(infile)            


    def loadLines(self, lines):
        """Loads data from a list of data lines."""
        if self.bVerbose: sys.stderr.write(" parsing...")
        ORFtoRow = {}
        for line in lines:
            if line[0] == '!':
                # skip comment lines, starting with '!'
                continue
            
            fields = line.split('\t')
            
            SGDID = fields[1]
            CGN = fields[2]
            NOT = fields[3]
            if NOT == 'NOT': continue
            GOID = fields[4]
            if GOID in self.ignoreNodes: continue
            
            assocType = fields[8]
            if assocType != 'P': continue # skip non-Process lines
            
            ORF = fields[10].split('|')[0] #truncate at "|" if any
            if self.funcTranslateORF is not None:
                t = self.funcTranslateORF(CGN)
                if t is None:
                    t = self.funcTranslateORF(ORF)
                if t is None:
                    if self.bVerbose:
                        sys.stderr.write('*** WARNING: Gene lost in translation "%s"\n' % ORF) ; continue
                    else:
                        pass
                else: ORF = t

            # add to data structures
            row = ORFtoRow.get(ORF)
            if row is not None:
                nodes = row['nodes']
                if GOID not in nodes:
                    nodes.append(GOID)
            else:
                row = [SGDID, CGN, ORF, [GOID], []]
                ORFtoRow[ORF] = self.genes.append(row)
        
        # save all annotations as direct annotations, because they change during propagation
        for g in self.genes:
            g['directNodes'] = g['nodes'][:]
        
        if self.bVerbose: sys.stderr.write("OK\n indexing...")
        self.genes.indexBy('SGDID', isUnique = True)
        self.genes.indexBy('CGN', isUnique = True)
        self.genes.indexBy('ORF', isUnique = True)
        self.genes.indexByList('nodes', isUnique = False)
        if self.bVerbose: sys.stderr.write("OK\n")
            
    def saveIndices(self, outfile, geneList = None, GOIDList = None, idcol = 'ORF'):
        """Saves data as sparse matrix text file.
        M(gene_index, GOID_index) = 1 if gene is associated to GOID.
        The indices are 1+(order in list parameter).
        If lists are not specified, order of appearance is used.
        """
        
        if geneList is None:
            geneList = self.genes.indexKeys(idcol)
        
        if GOIDList is None:
            GOIDList = self.genes.indexKeys('nodes')
        
        ftype = type(outfile)
        if ftype is str:
            f = file(outfile, 'wt')
        elif ftype is file:
            f = outfile
        
        try:
            # write items
            for (i, geneName) in zip(range(1,len(geneList)+1), geneList):
                nodesOfGene = self.getNodesOfGene(geneName, idcol)
                for (j, GOID) in zip(range(1,len(GOIDList)+1), GOIDList):
                    if GOID in nodesOfGene:
                        f.write('%d\t%d\t1\n' % (i, j))
            
            # ensure size by writing out item (m,n)
            lastGene = geneList[-1]
            lastGOID = GOIDList[-1]
            if lastGOID not in self.getNodesOfGene(lastGene, idcol):
                f.write('%d\t%d\t0\n' % (len(geneList), len(GOIDList)))
            
        finally:
            if ftype is str:
                f.close()

    def printSummary(self):
        print '%s summary:' % self.__class__.__name__
        print ' Num. genes = %d' % len(self.genes)
        print ' Num. nodes = %d' % len(self.genes.indexKeys('nodes'))
        print '(End)'

    def getNodesOfGene(self, geneName, idcol = 'ORF'):
        gene = self.genes.find(idcol, geneName)
        if gene is None: return []
        return gene['nodes']
        
    def getDirectNodesOfGene(self, geneName, idcol = 'ORF'):
        gene = self.genes.find(idcol, geneName)
        if gene is None: return []
        return gene['directNodes']

# Main test code
#assoc = association.Association() ; assoc.loadFile()
