"""
Gene Association data parser
Zafer Barutcuoglu 2003/11/25

This module parses the bioinformatical Gene Association data, which is the mapping of genes to Gene Ontology nodes.

The term "GOID" used throughout is short for Gene Ontology ID, of the format "GO:1234567".
"""

class Association:
    """Gene Association data class.
    
    self.genesToNodes is dict of gene_name -> list of GOID
    self.nodesToGenes is dict of GOID -> list of gene_name
    """
    
    def __init__(self, infile = None):
        """Constructor.
        <infile> may be either a file name, open file object, or None.
        """
        self.DEFAULT_FILE_NAME = 'C:\\Documents and Settings\\Zafer\\My Documents\\Machine Learning\\Olga\\Data\\GO\\gene_association.sgd.txt'

        self.genesToNodes = {}
        self.nodesToGenes = {}
        
        if infile is not None:
            self.loadFile(infile)


    def loadFile(self, infile = None):
        """Loads data from file.
        <infile> may be a file name or open file object. If latter, it is not closed on exit.
        """
        if infile is None:
            infile = self.DEFAULT_FILE_NAME

        ftype = type(infile)
        if ftype is str:
            f = file(infile, "rt")
        elif ftype is file:
            f = infile
        else:
            raise Exception("Parameter <infile> has invalid type: %s" % str(ftype))
        
        try:
            lines = f.readlines()
        finally:
            if ftype is str:
                f.close()
        
        self.loadLines(lines)


    def loadLines(self, lines):
        """Loads data from a list of data lines."""
        iLen = len(lines)
        iCur = 0
        
        while iCur < iLen:
            if lines[iCur][0] == '!':
                # skip comment lines, starting with '!'
                iCur += 1
                continue
            
            fields = lines[iCur].split('\t')
            
            GOID = fields[4]
            assocType = fields[8]
            geneName = fields[10].split('|')[0] #truncate at "|" if any
            # trim these, maybe?

            if assocType != 'F' or geneName == '':
                # skip non-Function or empty-gene-name lines
                iCur += 1
                continue
            
            # add to data structures
            
            if self.genesToNodes.has_key(geneName):
                self.genesToNodes[geneName].append(GOID)
            else:
                self.genesToNodes[geneName] = [GOID]
            
            if self.nodesToGenes.has_key(GOID):
                if geneName not in self.nodesToGenes[GOID]:
                    self.nodesToGenes[GOID].append(geneName)
            else:
                self.nodesToGenes[GOID] = [geneName]
            
            # next line
            iCur += 1
    
    
    def saveIndices(self, outfile, geneList = None, GOIDList = None):
        """Saves data as sparse matrix text file.
        M(gene_index, GOID_index) = 1 if gene is associated to GOID.
        The indices are 1+(order in list parameter).
        If lists are not specified, order of appearance is used.
        """
        
        if geneList is None:
            geneList = self.genesToNodes.keys()
        
        if GOIDList is None:
            GOIDList = self.nodesToGenes.keys()
        
        ftype = type(outfile)
        if ftype is str:
            f = file(outfile, 'wt')
        elif ftype is file:
            f = outfile
        
        try:
            # write items
            for (gene, GOIDs) in self.genesToNodes.iteritems():
                for GOID in GOIDs:
                    try:
                        i = geneList.index(gene)
                    except:
                        raise Exception("Gene '%s' not found in given gene list!" % gene)
                    try:
                        j = GOIDList.index(GOID)
                    except:
                        raise Exception("GOID '%s' not found in given GOID list!" % GOID)
                    f.write('%d\t%d\t1\n' % (i+1, j+1))
            
            # ensure size by writing out item (m,n)
            lastGOID = GOIDList[-1]
            lastGene = geneList[-1]
            if not self.genesToNodes.has_key(lastGene) or lastGOID not in self.genesToNodes[lastGene]:
                f.write('%d\t%d\t0\n' % (len(geneList), len(GOIDList)))
            
        finally:
            if ftype is str:
                f.close()

# Main test code
#assoc = association.Association() ; assoc.loadFile()
