"""
Gene Ontology Parser
Zafer Barutcuoglu 2003/10/03
This module parses the bioinformatical Gene Ontology (GO) data available at
http://www.geneontology.org/ontology/process.ontology into a tree structure.
"""

from TextData import TextData
from Table import Table
from sets import Set

import os

class Ontology(TextData):
    """Gene Ontology tree class.
    
    self.rootNode - Contains the multiple-inheritance tree after loading data.
    self.nodes - List of all unique tree nodes.
    self.nodesByID - Dictionary of nodes by GO ID string.
    self.association - Bound association object (annotations) if any
    """

    class Node:
        """Tree node class.
        
        self.name - Textual node name. NOT  unique!
        self.ids - List of GO ID strings that can identify this node.
        self.parents - list of ("<" or "%", parent_ref) tuples.
        self.children - list of ("<" or "%", child_ref) tuples.
        
        self.genes - list of gene objects, directly annotated or inherited.
        self.directGenes - list of gene objects, only directly annotated.
        self.ancestorGenes - list of gene objects, directly annotated to an ancestor (used for filtering N2 negatives)
        """
        
        def __init__(self, name, ids, parents = []):
            """name is string. id is string list. parents is list of ("<" or "%", parent_ref)."""
            
            self.name = name
            self.ids = ids
            self.parents = parents
            self.children = []
            self.genes = []
            self.directGenes = []
            self.ancestorGenes = []
        
        def annotate(self, gene, bDirect = True):
            if self.name == 'ROOT': return
            
            if bDirect:
                # put ONLY in self.directGenes; added to self.genes in separate call.
                if gene not in self.directGenes:
                    self.directGenes.append(gene)
            else:
                if gene not in self.genes:
                    self.genes.append(gene)
                    if len(Set(gene['nodes']).intersection(Set(self.ids))) == 0:
                        gene['nodes'].append(self.ids[0])
                    for (rel, p) in self.parents:
                        p.annotate(gene, False) # inherited annotation, as opposed to direct annotation
         
        def propagateAncestorGenes(self, parentAncestorGenes):
            # add genes from parent
            for g in parentAncestorGenes:
                if g not in self.ancestorGenes:
                    self.ancestorGenes.append(g)
            
            # send inherited+direct to children
            sendGenes = self.ancestorGenes[:]
            for g in self.directGenes:
                if g not in sendGenes:
                    sendGenes.append(g)            
            for (rel, childNode) in self.children:
                childNode.propagateAncestorGenes(sendGenes)
        
        def getParentsInSet(self, GOIDSet):
            """Returns the first ancestor in GOIDList along all parent paths. Each unique parent is returned once."""
            L = Set([])
            for (rel, p) in self.parents:
                pids = Set(p.ids).intersection(GOIDSet)
                if len(pids) > 0 and p not in L:
                    L.add(p)
                else:
                    pparents = p.getParentsInSet(GOIDSet)
                    L.union_update(pparents)
            return L
                
        def getAncestorsInSet(self, GOIDSet):
            """Returns all ancestors within GOIDSet."""
            ancestors = Set()
            for (rel, p) in self.parents:
                ancestors.union_update(p.getAncestorsInSet(GOIDSet))
                if len(Set(p.ids).intersection(GOIDSet)) > 0:
                    ancestors.add(p)
            return ancestors
                
        def getDirectParentsInSet(self, GOIDSet):
            """Returns the first DIRECT ancestor in GOIDList along all parent paths. Each unique parent is returned once.
            Ignores redundant GO edges.
            """
            # return parents - parent_ancestors
            setparents = Set()
            for (rel, p) in self.parents:
                if len(Set(p.ids).intersection(GOIDSet)) > 0:
                    setparents.add(p)
                else:
                    pparents = p.getParentsInSet(GOIDSet)
                    setparents.union_update(pparents)
            
            directparents = setparents.copy()
            for p in setparents:
                directparents.difference_update(p.getAncestorsInSet(GOIDSet))
            return directparents

    def __init__(self):
        """Constructor for an empty but valid tree."""
        self.DEFAULT_URL = 'http://www.geneontology.org/ontology/process.ontology'
        #self.DEFAULT_FILE_NAME = 'C:\\Documents and Settings\\Zafer\\My Documents\\Machine Learning\\Olga\\Data\\GO\\process.ontology.txt'
        self.DEFAULT_FILE_NAME = os.getenv('HOME') + '/research/function/rawdata/GO/process.ontology.txt'
        
        self.rootNode = self.Node('ROOT', [], [])
        self.nodes = []
        self.nodesByID = {}
        self.association = None

    def clear(self):
        """Clears all data, reverting to an empty tree."""
        del self.rootNode.children[:]
        del self.nodes[:]
        self.nodesByID.clear()
        self.annotated = False

    def loadLines(self, lines):
        """Loads an ontology tree from a list of strings."""
        self.loadLevel(lines, 0, 0, self.rootNode)


    def loadLevel(self, lines, iNextLine, level, parent):
        """Loads a subtree at <level> depth, starting at <iNextLine> in <lines>, under <parent>
        Returns the new value of iNextLine.
        """        
        newestNode = None # in correctly indented input data this should not be read before assigned.
        while iNextLine < len(lines):
            line = lines[iNextLine]
            
            # determine level of line
            thisLevel = 0
            while line[thisLevel] == ' ':
                thisLevel += 1
            
            if thisLevel < level:
                # end of subtree
                return iNextLine
            elif thisLevel > level:
                # recurse to deeper level under newest node.
                # anomaly if newestNode is unassigned.
                iNextLine = self.loadLevel(lines, iNextLine, thisLevel, newestNode)
            else:
                # if not a comment line,
                if line[level] != '!':
                    # add a new node under parent
                    newestNode = self.loadNode(line, level, parent, iNextLine)
                iNextLine += 1
            
        return iNextLine


    def loadNode(self, line, level, parent, lineNo = -1):
        """Creates a new node and adds it under parent, if not already existing.
        If already exists, only adds parent and child refs.
        Returns a ref to the node.
        """        
        # parse relation char, name and IDs
        relation = line[level]
        i = line.find(';', level)
        if i < 0:
            raise Exception('Line %d: Node ID not found' % lineNo)
        name = line[level+1:i].strip()
        j = i = i + 1
        while j < len(line) and line[j] not in "<!%;":
            j += 1
        ids = [id.strip() for id in line[i:j].split(',')]
        
        # try to find existing node
        for id in ids:
            node = self.nodesByID.get(id)
            if node is not None:
                break
        
        # if found, validate and add parent, otherwise create new.
        if node is not None:
            # found. check that everything matches
            if node.name != name:
                raise Exception('Line %d: Name %s conflicts with existing node %s;%s' % (lineNo, name, node.name, str(node.ids)))
            elif node.ids != ids:
                raise Exception('Line %d: IDs %s conflict with existing node %s;%s' % (lineNo, str(ids), node.name, str(node.ids)))
            
            # add upwards ref
            if (relation, parent) not in node.parents:
                node.parents.append( (relation, parent) )
        else:
            # not found. create a new node with one parent (upwards ref)
            node = self.Node(name, ids, [ (relation, parent) ])
            
            # add to global structures
            self.nodes.append(node)
            for id in ids:
                self.nodesByID[id] = node
        
        # add downwards ref
        if (relation, node) not in parent.children:
            parent.children.append( (relation, node) )
        
        return node

    def addAnnotations(self, assoc, bVerbose=True):
        missingNodes = Set()
        for gene in assoc.genes:
            # direct annotations
            genenodes = gene['directNodes']
            for nodename in genenodes:
                node = self.nodesByID.get(nodename)
                if node is None:
                    missingNodes.add(nodename)
                else:
                    node.annotate(gene, True)
            
            # inherited annotations
            genenodes = gene['nodes']
            for nodename in genenodes:
                node = self.nodesByID.get(nodename)
                if node is None:
                    missingNodes.add(nodename)
                else:
                    node.annotate(gene, False)
        
        self.rootNode.propagateAncestorGenes([])
        
        self.association = assoc
        if bVerbose and len(missingNodes) > 0:
            import sys
            sys.stderr.write('*** WARNING: Some annotated GO nodes not found:\n')
            for m in missingNodes:
                sys.stderr.write('  %s\n' % m)

    def saveLabels(self, outfile, ORFList, GOIDList):
        """Saves data as sparse matrix text file.
        M(gene_index, GOID_index) = 1 if gene is associated to GOID.
        The indices are 1+(order in list parameter).
        If lists are not specified, order of appearance is used.
        """
        
##        if geneList is None:
##            geneList = self.genes.indexKeys(idcol)
##        
##        if GOIDList is None:
##            GOIDList = self.genes.indexKeys('nodes')
##        
        ftype = type(outfile)
        if ftype is str:
            f = file(outfile, 'wt')
        elif ftype is file:
            f = outfile
        
        try:
            # write items
            for (j, GOID) in zip(range(1,len(GOIDList)+1), GOIDList):
                node = self.nodesByID[GOID]
                ancestorORFs = [g['ORF'] for g in node.ancestorGenes]
                positiveORFs = [g['ORF'] for g in node.genes]
                for (i, geneName) in zip(range(1,len(ORFList)+1), ORFList):
                    if geneName in ancestorORFs:
                        f.write('%d\t%d\tNaN\n' % (i, j))
                    elif geneName in positiveORFs:
                        f.write('%d\t%d\t1\n' % (i, j))
            
            # ensure size by writing out item (m,n)
            lastGene = ORFList[-1]
            lastGOID = GOIDList[-1]
            lastNode = self.nodesByID[lastGOID]
            ancestorORFs = [g['ORF'] for g in lastNode.ancestorGenes]
            positiveORFs = [g['ORF'] for g in lastNode.genes]
            if lastGene not in ancestorORFs and lastGene not in positiveORFs:
                f.write('%d\t%d\t0\n' % (len(ORFList), len(GOIDList)))
            
        finally:
            if ftype is str:
                f.close()

    def printSummary(self):
        print '%s summary:' % self.__class__.__name__
        print ' Num. nodes = %d' % len(self.nodes)
        if self.association is None:
            print ' Not annotated yet.'
            print '(End)'
        else:
            print ' Annotated:'
            self.association.printSummary()

    def exportNodeInfos(self, outfile, GOIDList, usedORFs):
        """Export information about selected nodes to Matlab ASCII file.
        For generating a web page (and possibly some stats) from Matlab.
        """
        ftype = type(outfile)
        if ftype is str:
            f = file(outfile, 'wt')
        elif ftype is file:
            f = outfile

        try:
            for goid in GOIDList:
                node = self.nodesByID[goid]
                
                lenUsedGenes = len([g for g in node.genes if g['ORF'] in usedORFs])
                f.write('"%s"\t"%s"\t%d\t%d\t%d\n' % (goid, node.name, len(node.directGenes), len(node.genes), lenUsedGenes))
            
        finally:
            if ftype is str:
                f.close()

    def exportNodeParents(self, outfile, GOIDList):
        """Export parent lists of selected nodes among the to Matlab ASCII file.
        For generating a web page (and possibly some stats) from Matlab.
        Parents here are the first ancestors in GOIDList along all parent paths; i.e., parents outside GOIDList are not returned.
        Each unique parent is returned only once.
        """
        ftype = type(outfile)
        if ftype is str:
            f = file(outfile, 'wt')
        elif ftype is file:
            f = outfile

        try:
            for goid in GOIDList:
                node = self.nodesByID[goid]
                pids = [list(Set(p.ids).intersection(Set(GOIDList)))[0] for p in node.getDirectParentsInSet(Set(GOIDList))]
                for pid in pids:
                    f.write('"%s"\t"%s"\n' % (pid, goid))

        finally:
            if ftype is str:
                f.close()

    def exportNodeParentsAsDOT(self, outfile, GOIDList, dictTexts, dictColors, dictProps, bShortLabel = False):
        """See exportNodeParents. Here the output is in GraphViz DOT file format.
        colorBy = {None | 'prob' | 'rank'}
        """
        if not bShortLabel:
            def makeLabel(goid, name, dictTexts):
                label = goid + '\\n' + name
                if dictTexts is not None:
                    text = dictTexts.get(goid)
                else:
                    text = None
                
                if text is not None:
                    label = label + '\\n' + text
                
                return label
        else:
            def makeLabel(goid, name, dictTexts):
                label = str(int(goid[4:]))
                return label
        
        ftype = type(outfile)
        if ftype is str:
            f = file(outfile, 'wt')
        elif ftype is file:
            f = outfile

        try:
            f.write('digraph {\n') # orientation="landscape";page="8.5,11";size="22.5,30";ranksep="5";\n')
            f.write('ranksep=1\n')
            f.write('node [fontname="Arial"]\n')
            
            # write colors if probs given            
            for goid in GOIDList:
                node = self.nodesByID[goid]
                props = ''
                
                if dictColors is not None:
                    rgbcode = dictColors.get(goid)
                    if rgbcode is not None:
                        props = props + 'style=filled,fillcolor="%s"' % rgbcode
                    
                if dictProps is not None:
                    props = props + dictProps.get(goid, '')
                        
                
                if len(props) > 0:
                    f.write('"%s" [%s]\n' % (makeLabel(goid, node.name, dictTexts), props))
                
                
            # write parent edges
            for goid in GOIDList:
                node = self.nodesByID[goid]
                pids = [list(Set(p.ids).intersection(Set(GOIDList)))[0] for p in node.getDirectParentsInSet(Set(GOIDList))]
                for pid in pids:
                    label = '"' + pid + '\\n' + self.nodesByID[pid].name + '"'
                    f.write('"%s" -> "%s"\n' % (makeLabel(pid, self.nodesByID[pid].name, dictTexts), \
                                                makeLabel(goid, node.name, dictTexts) ))
            
            f.write('}\n')
        finally:
            if ftype is str:
                f.close()

##    def exportNodeParentsAsDOT(self, outfile, GOIDList, dictProbs, dictRanks, colorBy):
##        """See exportNodeParents. Here the output is in GraphViz DOT file format.
##        colorBy = {None | 'prob' | 'rank'}
##        """
##        def makeLabel(goid, name, dictProbs, dictRanks):
##            label = goid + '\\n' + name
##            if dictProbs is not None:
##                prob = dictProbs.get(goid)
##            else:
##                prob = None
##            
##            if dictRanks is not None:
##                rank = dictRanks.get(goid)
##            else:
##                rank = None
##            
##            if prob is not None:
##                label = label + '\\np=%f' % prob
##            if rank is not None:
##                label = label + '\\nrank=%d' % rank
##            
##            return label
##            
##            
##        ftype = type(outfile)
##        if ftype is str:
##            f = file(outfile, 'wt')
##        elif ftype is file:
##            f = outfile
##
##        try:
##            f.write('digraph {\n') # orientation="landscape";page="8.5,11";size="22.5,30";ranksep="5";\n')
##            f.write('ranksep=1\n')
##            f.write('node [fontname="Arial"]\n')
##            
##            # write colors if probs given
##            if colorBy is not None and (dictProbs is not None):
##                for goid in GOIDList:
##                    node = self.nodesByID[goid]
##                    if colorBy == 'prob':
##                        prob = dictProbs.get(goid)
##                        if prob is None: continue
##                        if prob >= 0.5:
##                            hexcode = '00'
##                        else:
##                            hexcode = hex(int(255*(1-4*prob*prob)))[2:]
##                            hexcode = ('00'+hexcode)[-2:]
##                        rgbcode = '#FF%s%s' % (hexcode,hexcode)
##                        if prob >= 0.5:
##                            fontname = 'fontname="Arial-Bold",'
##                        else:
##                            fontname = ''
##                    elif colorBy == 'rank':
##                        rank = dictRanks.get(goid)
##                        if rank is None: continue
##                        if rank >= 50:
##                            hexcode = 'ff'
##                        else:
##                            hexcode = ('00' + hex(int(256.0*(rank-1)/50.0))[2:])[-2:]
##                        rgbcode = '#%s%sFF' % (hexcode,hexcode)
##                        if rank <= 10:
##                            fontname = 'fontname="Arial-Bold",'
##                        else:
##                            fontname = ''
##                    f.write('"%s" [%sstyle=filled,fillcolor="%s"]\n' % (makeLabel(goid, node.name, dictProbs, dictRanks), fontname, rgbcode))
##                
##                
##            # write parent edges
##            for goid in GOIDList:
##                node = self.nodesByID[goid]
##                pids = [list(Set(p.ids).intersection(Set(GOIDList)))[0] for p in node.getDirectParentsInSet(Set(GOIDList))]
##                for pid in pids:
##                    label = '"' + pid + '\\n' + self.nodesByID[pid].name + '"'
##                    f.write('"%s" -> "%s"\n' % (makeLabel(pid, self.nodesByID[pid].name, dictProbs, dictRanks), \
##                                                makeLabel(goid, node.name, dictProbs, dictRanks) ))
##            
##            f.write('}\n')
##        finally:
##            if ftype is str:
##                f.close()

    def exportNodeParentsAsDSL(self, outfile, GOIDList):
        """See exportNodeParents. Here the output is in GeNIe DSL file format."""
        ftype = type(outfile)
        if ftype is str:
            f = file(outfile, 'wt')
        elif ftype is file:
            f = outfile

        try:
            f.write('net N {\n')

            for goid in GOIDList:
                node = self.nodesByID[goid]
                pids = list(node.getParentsInSet(Set(GOIDList)))

                lenUsedGenes = len([g for g in node.genes if g['ORF'] in usedORFs])
                f.write('node %s {TYPE = CPT; HEADER = {NAME = "%s";}; PARENTS = ( %s );};\n' % (goid.replace(':','_'), goid, ",".join([pid.replace(':','_') for pid in pids])))
            
            f.write('};\n')
        finally:
            if ftype is str:
                f.close()

# go = ontology.Ontology(); go.loadFile('process.ontology.txt')
